from __future__ import absolute_import, division, print_function

import logging
import os
import re
import requests
import socket
from subprocess import STDOUT
try:
    from subprocess import SubprocessError
except ImportError:  # pragma: no cover
    # py2
    from subprocess import CalledProcessError as SubprocessError
import time
import warnings

from .utils import shell_out, get_log_content
from .exceptions import YARNException
logger = logging.getLogger(__name__)


class YARNAPI(object):
    """REST interface to YARN

    self.auth holds the authentication being used - it can be updated as
    needed, and will then be applied to subsequent actions.

    Parameters
    ----------
    rm: str
        Resource Manager host
    rm_port: int
        HTTP REST port
    scheme: 'http' or 'https'
    gateway_path: str
        If routing through a proxy, this is prepended to the URL path
    kerberos: bool
        Whether to use kerberos authentication when speaking to the REST
        service. Requires requests_kerberos, see
        https://github.com/requests/requests-kerberos
    username, password: str
        For simple authentication of the REST endpoint.
    """
    timeout = 2  # for REST HTTP calls

    def __init__(self, rm, rm_port, scheme='http', gateway_path='',
                 kerberos=False, username=None, password=None):
        self.rm = rm
        self.rm_port = rm_port
        self.scheme = scheme
        self.gateway_path = gateway_path
        self.host_port = "{0}:{1}".format(self.rm, self.rm_port)
        self.url = scheme + '://' + self.host_port + gateway_path + '/ws/v1/'
        if kerberos:
            from requests_kerberos import HTTPKerberosAuth
            self.auth = HTTPKerberosAuth()
        elif username and password:
            self.auth = (username, password)
        else:
            self.auth = None
        self.username = username
        self.password = password

    @property
    def apps(self):
        """App IDs known to YARN"""
        data = self.apps_info()
        apps = [d['id'] for d in data]
        return apps

    def apps_info(self, app_id=None):
        """List app apps, or info for given app"""
        if app_id is None:
            # this query allows for filtering on a number of parameters
            url = self.url + 'cluster/apps/'
            logger.debug("Getting Resource Manager Info: {0}".format(url))
            r = requests.get(url, timeout=self.timeout, auth=self.auth)
            self._verify_response(r)
            data = r.json()
            return (data.get('apps', None) or {'app': []})['app']
        else:
            r = requests.get(self.url + 'cluster/apps/{}'.format(app_id),
                             timeout=self.timeout, auth=self.auth)
            self._verify_response(r)
            return r.json()['app']

    def app_attempts(self, app_id):
        """List of attempt details for given app"""
        r = requests.get(self.url + 'cluster/apps/{}/appattempts'.format(
                app_id), timeout=self.timeout, auth=self.auth)
        self._verify_response(r)
        return r.json().get('appAttempts', {'app_attempt': []})['appAttempt']

    def app_containers(self, app_id=None, info=None):
        """
        Get list of container information for given app. If given app_id,
        will automatically get its info, or can skip by providing the info
        directly.
        
        Parameters
        ----------
        app_id: str
            YARN ID for the app
        info: dict
            Produced by app_info()
    
        Returns
        -------
        List of container info dictionaries
        """
        if (app_id is None) == (info is None):
            raise TypeError('Must provide app_id or info')
        if app_id:
            info = self.status(app_id)

        amHostHttpAddress = info['amHostHttpAddress']

        url = "http://{0}/ws/v1/node/containers".format(
            amHostHttpAddress)
        r = requests.get(url, timeout=self.timeout, auth=self.auth)
        self._verify_response(r)

        data = r.json()['containers']
        if not data:
            raise YARNException("No containers available")

        container = data['container']
        logger.debug(container)

        # container_1452274436693_0001_01_000001
        def get_app_id_num(x):
            return "_".join(x.split("_")[1:3])

        app_id_num = get_app_id_num(app_id)
        containers = [d for d in container
                      if get_app_id_num(d['id']) == app_id_num]
        return containers

    def logs(self, app_id, shell=False, retries=4, delay=3):
        """
        Collect logs from RM (if running)
        With shell=True, collect logs from HDFS after job completion

        Parameters
        ----------
        app_id: str
            A yarn application ID string
        shell: bool
            Shell out to yarn CLI (default False)
        retries: int
            If CLI is not returning info, retry this many times
        delay: number
            Seconds to wait between retries

        Returns
        -------
        log: dictionary
            logs from each container (when possible)
        """
        running = self.state(app_id) == 'RUNNING'
        if not shell and running:
            # logs are held in memory only while app is running
            try:
                containers = self.app_containers(app_id)
                logs = {}
                for c in containers:
                    log = dict(nodeId=c['nodeId'])

                    # grab stdout
                    url = "{0}/stdout/?start=0".format(c['containerLogsLink'])
                    logger.debug("Gather stdout/stderr data from {0}:"
                                 " {1}".format(c['nodeId'], url))
                    r = requests.get(url, timeout=self.timeout, auth=self.auth)
                    log['stdout'] = get_log_content(r.text)

                    # grab stderr
                    url = "{0}/stderr/?start=0".format(c['containerLogsLink'])
                    r = requests.get(url, timeout=self.timeout, auth=self.auth)
                    log['stderr'] = get_log_content(r.text)

                    logs[c['id']] = log
                return logs

            except Exception:
                logger.warning("Error while attempting to fetch logs,"
                               " using fallback", exc_info=1)

        # fallback
        # TODO: this is just a location in HDFS given by app info
        cmd = ["yarn", "logs", "-applicationId", app_id]
        while True:
            try:
                out = shell_out(cmd)
                break
            except SubprocessError:  # pragma: no cover
                retries -= 1
                if retries < 0:
                    raise RuntimeError('Retries exceeded when fetching logs for'
                                       ' ' + app_id)
                time.sleep(delay)
        logs = {}
        container = None
        ltype = 'stdout'
        started = False
        for line in out.split('\n'):
            p = re.compile('Container: ([a-zA-Z0-9_]+) on ([a-zA-Z0-9_]+)')
            r = p.match(line)
            if r:
                container, nodeID = r.groups()
                logs[container] = dict(nodeId=nodeID, stdout='', stderr='')
                started = False
            elif line == 'LogType:stderr':
                ltype = 'stderr'
            elif line == 'LogType:stdout':
                ltype = 'stdout'
            elif line == "Log Contents:":
                started = True
            elif started:
                logs[container][ltype] = logs[container][ltype] + '\n' + line

        return logs

    def container_status(self, container_id):
        """Ask the YARN shell about the given container

        Better to use app_containers, assuming you know the app_id, which
        is normally "_".join(container_id.split("_")[:3])
        """
        cmd = ["yarn", "container", "-status", container_id]
        return str(shell_out(cmd))

    def state(self, app_id):
        """Current state of given application"""
        r = requests.get(self.url + 'cluster/apps/{}/state'.format(app_id),
                         timeout=self.timeout, auth=self.auth)
        self._verify_response(r)
        return r.json()['state']
    
    def status(self, app_id):
        """ Get status of an application

        Parameters
        ----------
        app_id: str
             A yarn application ID string

        Returns
        -------
        dictionary: status of application
        """
        return self.apps_info(app_id)

    def kill_all(self, knit_only=True):
        """Kill a set of applications
        
        Parameters
        ----------
        knit_only: bool (True)
            Only kill apps with the name 'knit' (i.e., ones we started)
        """
        for app in self.apps:
            stat = self.apps_info(app)
            if knit_only and stat['name'] != 'knit':
                continue
            if stat['state'] in ['KILLED', 'FINISHED', 'FAILED']:
                continue
            self.kill(app)

    def kill(self, app_id):
        """
        Method to kill a yarn application

        Parameters
        ----------
        app_id: str
            YARN application id

        Returns
        -------
        bool:
            True if successful, False otherwise.
        """

        cmd = ["yarn", "application", "-kill", app_id]
        try:
            out = shell_out(cmd, stderr=STDOUT)
            return "Killed application" in out
        except SubprocessError:
            return False

    def _verify_response(self, r):
        if not r.ok:
            try:
                ex = r.json()['RemoteException']
                raise YARNException(ex.get('message', str(ex)))
            except (ValueError, IndexError):
                raise YARNException(r.text)

    def cluster_info(self):
        """YARN cluster information: driver, version..."""
        r = requests.get(self.url + 'cluster', timeout=self.timeout,
                         auth=self.auth)
        self._verify_response(r)
        return r.json()['clusterInfo']

    def cluster_metrics(self):
        """YARN cluster global capacity/allocations"""
        r = requests.get(self.url + 'cluster/metrics', timeout=self.timeout,
                         auth=self.auth)
        self._verify_response(r)
        return r.json()['clusterMetrics']

    def scheduler(self):
        """State of the scheduler/queue"""
        r = requests.get(self.url + 'cluster/scheduler', timeout=self.timeout,
                         auth=self.auth)
        self._verify_response(r)
        return r.json()['scheduler']

    def app_stats(self):
        """Number of apps of various states"""
        r = requests.get(self.url + 'cluster/appstatistics',
                         timeout=self.timeout, auth=self.auth)
        self._verify_response(r)
        return r.json()['appStatInfo']

    def nodes(self):
        """Info on YARN's worker nodes"""
        r = requests.get(self.url + 'cluster/nodes', timeout=self.timeout,
                         auth=self.auth)
        self._verify_response(r)
        return r.json()['nodes']['node']

    def system_logs(self):
        """Trouble-shooting method

        This will check for RM and NM processes on this machine, and return
        their logs, if possible.
        """
        nodes = self.nodes()
        ips = set((socket.gethostbyname_ex(socket.gethostname())[2] +
                   socket.gethostbyname_ex('localhost')[2]))
        on_rm = socket.gethostbyname(self.rm) in ips
        single_node = on_rm and len(nodes) == 1 and (
            socket.gethostbyname(nodes[0]['nodeHostName']) in ips)
        out = {'single_node': single_node, 'on_rm': on_rm}
        try:
            import psutil
            for p in psutil.process_iter():
                if any('NodeManager' in s for s in p.cmdline()):
                    out['nm_proc'] = p
                    log = os.path.join([s.split('=')[1] for s in p.cmdline()
                                        if s.startswith('-Dyarn.log.dir')][0],
                                       [s.split('=')[1] for s in p.cmdline()
                                        if s.startswith('-Dyarn.log.file')][0])
                    out['nm_logfile'] = log
                if any('ResourceManager' in s for s in p.cmdline()):
                    out['rm_proc'] = p
                    log = os.path.join([s.split('=')[1] for s in p.cmdline()
                                        if s.startswith('-Dyarn.log.dir')][0],
                                       [s.split('=')[1] for s in p.cmdline()
                                        if s.startswith('-Dyarn.log.file')][0])
                    out['rm_logfile'] = log
        except ImportError:
            warnings.warn('psutil is not installed')
        return out



from __future__ import absolute_import, division, print_function
import warnings

from .utils import *
from .core import *
from .env import CondaCreator, zip_path
from .yarn_api import YARNAPI


try:
    from dask_yarn import DaskYARNCluster as _DaskYARNCluster

    class DaskYARNCluster(_DaskYARNCluster):
        def __init__(self, *args, **kwargs):
            warnings.warn("DeprecationWarning: `knit.DaskYARNCluster` is "
                          "deprecated, please use "
                          "`dask_yarn.DaskYARNCluster` instead")
            super().__init__(*args, **kwargs)
except ImportError:
    pass

__version__ = "0.2.4"



"""
Configuration from system files and environment variables.

Warning: much of this code is copied from hdf3.conf
"""

from __future__ import absolute_import

import os
import re
try:
    from urllib.parse import urlsplit
except ImportError:
    from urlparse import urlsplit


__all__ = ('get_config', 'load_config', 'reset_config_cache')


def get_config(autodetect=True, pars=None, **kwargs):
    """Get configuration dictionary.

    Build up a configuration dictionary based on configuration files and
    overrides.

    Parameters
    ----------
    autodetect : bool, optional
        If True (default) configuration is read from files on disk. Otherwise
        only specified configuration parameters are used.
    pars : dict, optional
        Extra parameters to use for overriding
    **kwargs
        Extra parameters to use for overriding
    """
    config = load_config() if autodetect else {}
    if pars:
        config.update(**pars)
    config.update(**kwargs)
    return config


def current_user():
    try:
        import pwd
        return pwd.getpwuid(os.geteuid()).pw_name
    except ImportError:
        import getpass
        return getpass.getuser()


java_lib_dir = os.path.join(os.path.dirname(__file__), "java_libs")
DEFAULT_KNIT_HOME = os.environ.get('KNIT_HOME') or java_lib_dir

# standard defaults
DEFAULTS = {'rm': 'localhost',
            'rm_port': 8088,
            'rm_port_https': 8090,
            'replication_factor': 3,
            'user': current_user()}


def get_host_port(addr):
    """Infer a host & port from a given addr"""
    parsed = urlsplit(addr)
    if parsed.hostname:
        host = parsed.hostname
        port = parsed.port
    elif ':' in parsed.path:
        host, port = parsed.path.split(':', 1)
        port = int(port)
    else:
        host = parsed.path
        port = None
    return host, port


def infer_extra_params(config):
    """Given a config dict, infer the values of some extra fields"""
    user = current_user()

    replication_factor = config.get('dfs.replication',
                                    DEFAULTS['replication_factor'])

    # resourcemanager and port
    if 'yarn.resourcemanager.webapp.address' in config:
        rm_addr = config['yarn.resourcemanager.webapp.address']
        rm, rm_port = get_host_port(rm_addr)
    else:
        rm = config.get('yarn.resourcemanager.hostname', DEFAULTS['rm'])
        rm_port = DEFAULTS['rm_port']

    # resourcemanager https port
    if 'yarn.resourcemanager.webapp.https.address' in config:
        rm_https_addr = config['yarn.resourcemanager.webapp.https.address']
        rm_port_https = get_host_port(rm_https_addr)[1]
    else:
        rm_port_https = DEFAULTS['rm_port_https']

    return {'user': user,
            'replication_factor': replication_factor,
            'rm': rm,
            'rm_port': rm_port,
            'rm_port_https': rm_port_https}


def config_to_dict(fname):
    """ Read a *-site.xml style conf file, produces dictionary """
    name_match = re.compile("<name>(.*?)</name>")
    val_match = re.compile("<value>(.*?)</value>")
    conf = {}
    for line in open(fname):
        name = name_match.search(line)
        if name:
            key = name.groups()[0]
        val = val_match.search(line)
        if val:
            val = val.groups()[0]
            conf[key] = val
    return conf


def error_if_path_missing(envvar):
    """Raise a ValueError if the path specified by an ENVVAR is missing"""
    path = os.environ[envvar]
    if not os.path.exists(path):
        raise ValueError("Environment variable `%s` points to "
                         "non-existant location" % envvar)


def find_config_files():
    """Look for config files in common places"""

    # Find the configuration directory
    if 'LIBHDFS3_CONF' in os.environ:
        error_if_path_missing('LIBHDFS3_CONF')
        confd, hdfs_site = os.path.split(os.environ['LIBHDFS3_CONF'])
    elif 'HADOOP_CONF_DIR' in os.environ:
        error_if_path_missing('HADOOP_CONF_DIR')
        confd = os.environ['HADOOP_CONF_DIR']
    elif 'HADOOP_INSTALL' in os.environ:
        error_if_path_missing('HADOOP_INSTALL')
        confd = os.environ['HADOOP_INSTALL'] + '/hadoop/conf'
    else:
        # list of potential typical system locations
        for loc in ['/etc/hadoop/conf']:
            if os.path.exists(loc):
                fns = os.listdir(loc)
                if 'core-site.xml' in fns:
                    confd = loc
                    break
        else:
            # fallback: local dir
            confd = os.getcwd()

    configs = {'yarn': 'yarn-site.xml',
               'core': 'core-site.xml'}
    paths = {}

    for name, f in configs.items():
        full_path = os.path.join(confd, f)
        if not os.path.exists(full_path):
            raise ValueError("Failed to load `%s` configuration file at "
                             "`%s`, file not found" % (name, full_path))
        paths[name] = full_path

    return paths


# The cached configuration
_cached_config = None


def load_config():
    """Load configuration from config files.

    Will error if configuration files aren't found.

    Returns
    -------
    config : dict
        The configuration dictionary.
    """
    global _cached_config

    if _cached_config is None:
        # Find the configuration files
        try:
            paths = find_config_files()
        except ValueError:
            _cached_config = DEFAULTS.copy()
            return DEFAULTS.copy()

        # Load and merge all config files
        config = {}
        for path in paths.values():
            config.update(config_to_dict(path))

        # Determine extra parameters
        config.update(infer_extra_params(config))

        _cached_config = config

    return _cached_config.copy()


def reset_config_cache():
    """Reset the configuration cache.

    The configuration will be recomputed on the next call to `load_config`."""
    global _cached_config
    _cached_config = None



from __future__ import absolute_import, division, print_function

import atexit
import os
import logging
import socket
import select
import signal
import platform
import requests
import socket
from subprocess import Popen, PIPE, call
import struct
import time
import weakref

from .conf import get_config, DEFAULT_KNIT_HOME
from .env import CondaCreator
from .exceptions import KnitException, YARNException
from .yarn_api import YARNAPI
from .utils import triple_slash

from py4j.protocol import Py4JError
from py4j.java_gateway import JavaGateway, GatewayClient
from py4j.java_collections import MapConverter, ListConverter

logger = logging.getLogger(__name__)
on_windows = platform.system() == "Windows"


def read_int(stream):
    length = stream.read(4)
    if not length:
        raise EOFError
    return struct.unpack("!i", length)[0]


class Knit(object):
    """
    Connection to HDFS/YARN. Launches a single "application" master with a
    number of worker containers.
    
    Parameter definition (nn, nn_port, rm, rm_port): those parameters given
    to __init__ take priority. If autodetect=True, Knit will attempt to fill
    out the others from system configuration files; fallback values are provided
    if this fails.

    Parameters
    ----------
    nn: str
        Namenode hostname/ip
    nn_port: int
        Namenode Port (default: 9000)
    rm: str
        Resource Manager hostname
    rm_port: int
        Resource Manager port (default: 8088)
    lang: str
        Environment variable language setting, required for ``click`` to
        successfully read from the shell. (default: 'C.UTF-8')
    user: str ('root')
        The user name from point of view of HDFS. This is only used when
        checking for the existence of knit files on HDFS, since they are stored
        in the user's home directory.
    hdfs_home: str
        Explicit location of a writable directory in HDFS to store files. 
        Defaults to the user 'home': hdfs://user/<username>/
    replication_factor: int (3)
        replication factor for files upload to HDFS (default: 3)
    autodetect: bool
        Autodetect configuration
    upload_always: bool(=False)
        If True, will upload conda environment zip always; otherwise will
        attempt to check for the file's existence in HDFS (using the hdfs3
        library, if present) and not upload if that matches the existing local
        file in size and is newer.
    knit_home: str
        Location of knit's jar
    hdfs: HDFileSystem instance or None
        Used for checking files in HDFS.

    Note: for now, only one Knit instance can live in a single process because
    of how py4j interfaces with the JVM.

    Examples
    --------

    >>> k = Knit()
    >>> app_id = k.start('sleep 100', num_containers=5, memory=1024)
    """

    JAR_FILE = "knit-1.0-SNAPSHOT.jar"
    JAVA_APP = "io.continuum.knit.Client"
    _instances = weakref.WeakSet()

    def __init__(self, autodetect=True, upload_always=False, hdfs_home=None,
                 knit_home=DEFAULT_KNIT_HOME, hdfs=None, pars=None,
                 **kwargs):

        self.conf = get_config(autodetect=autodetect, pars=pars, **kwargs)
        gateway_path = self.conf.get('gateway_path', '')
        kerb = self.conf.get(
            'hadoop.http.authentication.type', '') == 'kerberos'
        if not kerb and self.conf.get('hadoop.http.authentication.simple.'
                                      'anonymous.allowed', '') == 'false':
            if 'password' not in self.conf:
                raise KnitException('Simple auth required: please supply'
                                    '`password=`.')
            pw = self.conf['password']
        else:
            pw = None

        if self.conf.get('yarn.http.policy', '').upper() == "HTTPS_ONLY":
            self.yarn_api = YARNAPI(self.conf['rm'], self.conf['rm_port_https'],
                                    scheme='https', gateway_path=gateway_path,
                                    kerberos=kerb, username=self.conf['user'],
                                    password=pw)
        else:
            self.yarn_api = YARNAPI(self.conf['rm'], self.conf['rm_port'],
                                    gateway_path=gateway_path,
                                    kerberos=kerb, username=self.conf['user'],
                                    password=pw)

        self.KNIT_HOME = knit_home
        self.upload_always = upload_always
        self.lang = self.conf.get('lang', 'C.UTF-8')
        self.hdfs_home = hdfs_home or self.conf.get(
            'dfs.user.home.base.dir', '/user/' + self.conf['user'])
        self.client_gateway = None

        # must set KNIT_HOME ENV for YARN App
        os.environ['KNIT_HOME'] = self.KNIT_HOME
        os.environ['REPLICATION_FACTOR'] = str(self.conf['replication_factor'])
        os.environ['HDFS_KNIT_DIR'] = self.hdfs_home

        self.client = None
        self.master = None
        self.app_id = None
        self.proc = None
        self.hdfs = hdfs
        self._instances.add(self)

    def __repr__(self):
        return "Knit<RM={0}:{1}>".format(self.conf['rm'], self.conf['rm_port'])

    @property
    def JAR_FILE_PATH(self):
        return os.path.join(self.KNIT_HOME, self.JAR_FILE)

    def _pre_flight_checks(self, num_containers, virtual_cores, memory,
                           files, queue):
        """Some checks to see if app is possible to schedule

        This depends on YARN's allocations reporting, which do not necessarily
        reflect the true amount of resources on the cluster. Other failure
        modes, such as full disc, are not likely to be caught here.
        """
        try:
            # check response from RM
            met = self.yarn_api.cluster_metrics()
        except YARNException:
            raise
        except requests.RequestException as e:
            if isinstance(e, requests.Timeout):
                m = 'Connection timeout'
            else:
                m = 'Connection error'
            raise YARNException(m + ' when talking to the '
                                'YARN REST server at {}. This can mean that '
                                'the server/port values are wrong, that you '
                                'are using the wrong protocol (http/https) or '
                                'that you need to route through a proxy.'
                                ''.format(self.yarn_api.url))
        if met['activeNodes'] < 1:
            raise KnitException('No name-nodes active')
        # What if we simply don't have the full yarn-site.xml available?
        mmin = int(self.conf.get('yarn.scheduler.minimum-allocation-mb', 1024))
        # 300MB default allocation for AM in client.scala
        mem = (max(300, mmin) + num_containers * max(memory, mmin))
        if met['availableMB'] < mem:
            raise KnitException('Memory estimate for app (%iMB) exceeds cluster'
                                ' capacity (%iMB)' % (mem, met['availableMB']))
        c = 1 + num_containers * virtual_cores
        if met['availableVirtualCores'] < c:
            raise KnitException('vCPU request for app (%i) exceeds cluster capa'
                                'city (%i)' % (c, met['availableVirtualCores']))
        nodes = self.yarn_api.nodes()
        if all((max(mmin, memory) > n['availMemoryMB']) and
               (virtual_cores > n['availableVirtualCores'])
               for n in nodes):
            # cannot test without multiple nodemanagers
            raise KnitException('No NodeManager can fit any single container')
        if self.hdfs:
            df = self.hdfs.df()
            cap = (df['capacity'] - df['used']) // 2**20
            fs = [self.JAR_FILE_PATH] + [f for f in files
                                         if not f.startswith('hdfs://')]
            need = sum(os.stat(f).st_size for f in fs) // 2**20
            # NB: if replication > 1 this might not be enough
            if cap < need:
                raise KnitException('HDFS space requirement (%iMB) exceeds'
                                    'capacity (%iMB)' % (need, cap))

    def start(self, cmd, num_containers=1, virtual_cores=1, memory=128,
              files=None, envvars=None, app_name="knit", queue="default",
              checks=True):
        """
        Method to start a yarn app with a distributed shell

        Parameters
        ----------
        cmd: str
            command to run in each yarn container
        num_containers: int
            Number of containers YARN should request (default: 1)
            * A container should be requested with the number of cores it can
              saturate, i.e.
            * the average number of threads it expects to have runnable at a
              time.
        virtual_cores: int
            Number of virtual cores per container (default: 1)
            * A node's capacity should be configured with virtual cores equal to
            * its number of physical cores.
        memory: int
            Memory per container (default: 128)
            * The unit for memory is megabytes.
        files: list
            list of files to be include in each container. If starting with
            `hdfs://`, assume these already exist in HDFS and don't need
            uploading. Otherwise, if hdfs3 is installed, existence of the
            file on HDFS will be checked to see if upload is needed.
            Files ending with `.zip` will be decompressed in the
            container before launch as a directory with the same name as the
            file: if myarc.zip contains files inside a directory stuff/, to
            the container they will appear at ./myarc.zip/stuff/* .
        envvars: dict
            Environment variables to pass to AM *and* workers. Both keys
            and values must be strings only.
        app_name: String
            Application name shown in YARN (default: "knit")
        queue: String
            RM Queue to use while scheduling (default: "default")
        checks: bool=True
            Whether to run pre-flight checks before submitting app to YARN

        Returns
        -------
        applicationId: str
            A yarn application ID string
        """
        files = files or []
        envvars = envvars or {'KNIT_LANG': self.lang}
        for k, v in envvars.items():
            if not isinstance(k, str) or not isinstance(v, str):
                raise ValueError('Environment must contain only strings (%s)'
                                 % ((k, v),))
        if self.app_id:
            raise ValueError('Already started')
        if not isinstance(memory, int):
            raise KnitException("Memory argument must be an integer")
        if files:
            if not isinstance(files, list):
                raise KnitException("File argument must be a list of strings")

        if checks:
            self._pre_flight_checks(num_containers, virtual_cores, memory,
                                    files, queue)
        # From https://github.com/apache/spark/blob/d83c2f9f0b08d6d5d369d9fae04cdb15448e7f0d/python/pyspark/java_gateway.py
        # thank you spark

        ## Socket for PythonGatewayServer to communicate its port to us
        callback_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        callback_socket.bind(('127.0.0.1', 0))
        callback_socket.listen(1)
        callback_host, callback_port = callback_socket.getsockname()

        if not os.path.exists(self.JAR_FILE_PATH):
            raise KnitException('JAR file %s does not exists - please build'
                                ' with maven' % self.JAR_FILE_PATH)
        args = ["hadoop", "jar", self.JAR_FILE_PATH, self.JAVA_APP,
                "--callbackHost", str(callback_host), "--callbackPort",
                str(callback_port)]

        ## Launch the Java gateway.
        # We open a pipe to stdin so that the Java gateway can die when the pipe is broken
        if not on_windows:
            # Don't send ctrl-c / SIGINT to the Java gateway:
            def preexec_func():
                signal.signal(signal.SIGINT, signal.SIG_IGN)
            proc = Popen(args, stdin=PIPE, preexec_fn=preexec_func)
        else:
            # preexec_fn not supported on Windows
            proc = Popen(args, stdin=PIPE)
        self.proc = proc
        gateway_port = None
        # We use select() here in order to avoid blocking indefinitely if the
        # subprocess dies before connecting
        long_timeout = 60
        while gateway_port is None and proc.poll() is None and long_timeout > 0:
            timeout = 1  # (seconds)
            readable, _, _ = select.select([callback_socket], [], [], timeout)
            if callback_socket in readable:
                gateway_connection = callback_socket.accept()[0]
                # Determine which ephemeral port the server started on:
                gateway_port = read_int(gateway_connection.makefile(mode="rb"))
                gateway_connection.close()
                callback_socket.close()
            long_timeout -= 1

        if gateway_port is None:
            raise Exception("The JVM Knit client failed to launch successfully."
                            " Check that java is installed and the Knit JAR"
                            " file exists.")

        gateway = JavaGateway(GatewayClient(port=gateway_port),
                              auto_convert=True)
        self.client = gateway.entry_point
        self.client_gateway = gateway
        logger.debug("Files submitted: %s" % files)
        upfiles = [f for f in files if (not f.startswith('hdfs://')
                   and self.check_needs_upload(f))]
        logger.debug("Files to upload: %s" % upfiles)
        jfiles = ListConverter().convert(upfiles, gateway._gateway_client)
        jenv = MapConverter().convert(envvars, gateway._gateway_client)

        self.app_id = self.client.start(jfiles, jenv, app_name, queue)

        ## Wait for AM to appear
        long_timeout = 100
        master_rpcport = -1
        while master_rpcport == -1:
            master_rpcport = self.client.masterRPCPort()
            time.sleep(0.2)
            long_timeout -= 0.2
            if long_timeout < 0:
                break

        if master_rpcport in [-1, 'N/A']:
            raise Exception(
"""The application master JVM process failed to report back. This can mean:
 - that the YARN cluster cannot scheduler adequate resources - check
   k.yarn_api.cluster_metrics() and other diagnostic methods;
 - that the ApplicationMaster crashed - check the application logs, k.logs();
 - that the cluster is otherwise unhealthy - check the RM and NN logs 
   (use k.yarn_api.system_logs() to find these on a one-node system
""")
        master_rpchost = self.client.masterRPCHost()

        gateway = JavaGateway(GatewayClient(
            address=master_rpchost, port=master_rpcport), auto_convert=True)
        self.master = gateway.entry_point
        rfiles = [triple_slash(f) if f.startswith('hdfs://') else
                  '/'.join(['hdfs://', self.hdfs_home, '.knitDeps',
                            os.path.basename(f)])
                  for f in files]
        logger.debug("Resource files: %s" % rfiles)
        jfiles = ListConverter().convert(rfiles, gateway._gateway_client)
        jenv = MapConverter().convert(envvars, gateway._gateway_client)
        self.master.init(jfiles, jenv, cmd, num_containers,
                         virtual_cores, memory)

        return self.app_id

    def add_containers(self, num_containers=1, virtual_cores=1, memory=128):
        """
        Method to add containers to an already running yarn app

        num_containers: int
            Number of containers YARN should request (default: 1)
            * A container should be requested with the number of cores it can
              saturate, i.e.
            * the average number of threads it expects to have runnable at a
              time.
        virtual_cores: int
            Number of virtual cores per container (default: 1)
            * A node's capacity should be configured with virtual cores equal to
            * its number of physical cores.
        memory: int
            Memory per container (default: 128)
            * The unit for memory is megabytes.
        """
        self.master.addContainers(num_containers, virtual_cores, memory)

    def get_containers(self):
        """
        Method to return active containers

        Returns
        -------
        container_list: List
            List of dicts with each container's details

        """
        if self.app_id:
            return self.yarn_api.app_containers(self.app_id)
        else:
            raise KnitException('Cannot get containers, app has not started')

    def get_container_statuses(self):
        """Get status info for each container

        Returns dict where the values are the raw text output.
        """
        return {c['id']: c['state'] for c in self.get_containers()}

    def remove_containers(self, container_id):
        """
        Method to remove containers from a running yarn app

            Calls removeContainers in ApplicationMaster.scala

        Be careful removing the ...0001 container.  This is where the
        applicationMaster is running

        Parameters
        ----------
        container_id: str

        Returns
        -------
        None

        """
        if container_id not in self.get_container_statuses():
            raise KnitException('Attempt to remove container nor owned by this'
                                'app: ' + container_id)
        self.master.removeContainer(str(container_id))

    @staticmethod
    def create_env(env_name, packages=None, remove=False,
                   channels=None, conda_pars=None):
        """
        Create zipped directory of a conda environment

        Parameters
        ----------
        env_name : str
        packages : list
        conda_root: str
            Location of conda installation. If None, will download miniconda and
            produce an isolated environment.
        remove : bool
            remove possible conda environment before creating
        channels : list of str
            conda channels to use (defaults to your conda setup)
        conda_pars: dict
            Further pars to pass to CondaCreator

        Returns
        -------
        path: str
            path to zipped conda environment

        Examples
        --------

        >>> k = Knit()
        >>> pkg_path = k.create_env(env_name='dev',
        ...                         packages=['distributed', 'dask', 'pandas'])
        """

        channels = channels or []
        c = CondaCreator(channels=channels, **(conda_pars or {}))
        return c.create_env(env_name, packages=packages, remove=remove)

    def logs(self, shell=False):
        """
        Collect logs from RM (if running)
        With shell=True, collect logs from HDFS after job completion

        Parameters
        ----------
        shell: bool
             Shell out to yarn CLI (default False)

        Returns
        -------
        log: dictionary
            logs from each container (when possible)
        """
        if self.app_id:
            return self.yarn_api.logs(self.app_id, shell=shell)
        else:
            raise KnitException('Cannot get logs, app not started')

    def print_logs(self, shell=False):
        """print out a more console-friendly version of logs()"""
        for l, v in self.logs(shell).items():
            print('\n### Container ', l, ', id ', v.get('id', 'None'), ' ###\n')
            for part in ['stdout', 'stderr']:
                print('##', part, '##')
                print(v[part])

    def wait_for_completion(self, timeout=10):
        """
        Wait for completion of the yarn application
        
        Returns
        -------
        bool:
            True if successful, False otherwise
        """
        cur_status = self.runtime_status()
        while cur_status not in ['FAILED', 'KILLED', 'FINISHED']:
            time.sleep(0.2)
            timeout -= 0.2
            cur_status = self.runtime_status()
            if timeout < 0:
                break

        return timeout > 0

    def kill(self):
        """
        Method to kill a yarn application

        Returns
        -------
        bool:
            True if successful, False otherwise.
        """
        if self.client is None:
            # never started, can't stop - should be warning or exception?
            return False
        try:
            self.client.kill()
        except Py4JError:
            logger.debug("Error while attempting to kill", exc_info=1)
            # fallback
            self.yarn_api.kill(self.app_id)
        if self.proc is not None:
            self.client_gateway.shutdown()
            if on_windows:
                call(["cmd", "/c", "taskkill", "/f", "/t", "/pid",
                      str(self.proc.pid)])
            self.proc.terminate()
            self.proc.communicate()
            self.proc = None
        self.client = None
        out = self.runtime_status() == 'KILLED'
        return out

    def __del__(self):
        if self.app_id is not None:
            try:
                self.kill()
            except:
                pass
        self.app_id = None

    def status(self):
        """ Get status of an application

        Returns
        -------
        log: dictionary
            status of application
        """
        if self.app_id:
            return self.yarn_api.apps_info(self.app_id)
        else:
            raise KnitException("Cannot get status, app not started")
    
    def runtime_status(self):
        """ Get runtime status of an application

        Returns
        -------
        str:
            status of application
        """
        try:
            return self.yarn_api.state(self.app_id)
        except:
            return "NONE"

    def list_envs(self):
        """List knit conda environments already in HDFS
        
        Looks in staging directory for zip-files
        
        Returns: list of dict
            Details for each zip-file."""
        if self.hdfs:
            files = self.hdfs.ls(self.hdfs_home + '/.knitDeps/', True)
            return [f for f in files if f['name'].endswith('.zip')]
        else:
            raise ImportError('Set the `hdfs` attribute to be able to list'
                              'environments.')

    def check_needs_upload(self, path):
        """Upload is needed if file does not exist in HDFS or is older"""
        if self.upload_always:
            return True
        fn = '/'.join([self.hdfs_home, '.knitDeps', os.path.basename(path)])
        if self.hdfs and self.hdfs.exists(fn):
            st = os.stat(path)
            size = st.st_size
            t = st.st_mtime
            info = self.hdfs.info(fn)
            if info['size'] == size and t < info['last_mod']:
                return False
            else:
                return True
        else:
            return True

    @classmethod
    def _cleanup(cls):
        # called on program exit to destroy lingering connections/apps
        for instance in cls._instances:
            instance.kill()

atexit.register(Knit._cleanup)



from __future__ import print_function, division, absolute_import

import logging

from .compatibility import check_output

format = ('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
logging.basicConfig(format=format, level=logging.INFO)
logging.getLogger("requests").setLevel(logging.WARNING)


def set_logging(level):
    logger = logging.getLogger('knit')
    logger.setLevel(level)

set_logging('INFO')


def shell_out(cmd=None, **kwargs):
    """
    Thin layer on check_output to return data as strings

    Parameters
    ----------
    cmd : list
        command to run
    kwargs:
        passed directly to check_output

    Returns
    -------
    result : str
        result of shell command
    """
    return check_output(cmd, **kwargs).decode('utf-8')


def get_log_content(s):
    if 'Cannot find this log' in s:
        return ''
    st = """<td class="content">"""
    ind0 = s.find(st) + len(st)
    ind1 = s[ind0:].find("</td>")
    out = s[ind0:ind0+ind1]
    return out.lstrip('\n          <pre>').rstrip('</pre>\n        ')


def triple_slash(s):
    if s.startswith('hdfs://') and not s.startswith('hdfs:///'):
        return 'hdfs:///' + s[7:]
    else:
        return s



import os
import re
import socket
import atexit
from hashlib import sha1

from tornado import gen
from toolz import unique

from knit import Knit, CondaCreator, zip_path
from distributed import LocalCluster

global_packages = ['dask>=0.14', 'distributed>=1.16']

prog = re.compile('\w+')


def first_word(s):
    return prog.match(s).group()


class DaskYARNCluster(object):
    """
    Implements a dask cluster with YARN containers running the worker processes.
    A dask scheduler is started locally upon instantiation, but you must call
    ``start()`` to initiate the building of containers by YARN.
    
    Parameters
    ----------
    nn, nn_port, rm, rm_port, user, autodetect, lang: see knit.Knit
    env: str or None
        If provided, the path of a zipped conda env to put in containers. This
        can be a local zip file to upload, a zip file already on HDFS (hdfs://)
        or a directory to zip and upload. If not provided, a default environment
        will be built, containing dask.
    packages: list of str
        Packages to install in the env to provide to containers *if* env is 
        None. Uses conda spec for pinning versions. dask and distributed will
        always be included.
    channels: list of str
        If building an environment, pass these extra channels to conda using
        ``-c`` (i.e., in addition but of superior priority to any system
        default channels).
    conda_pars: dict
        Things to pass to CondaCreator
    ip: IP-like string or None
        Address for the scheduler to listen on. If not given, uses the system
        IP.
    """

    def __init__(self, autodetect=True, packages=None, ip=None, env=None,
                 channels=None, conda_pars=None, **kwargs):

        ip = ip or socket.gethostbyname(socket.gethostname())

        self.env = env
        self.application_master_container = None
        self.app_id = None
        self.channels = channels or []
        self.conda_pars = conda_pars

        try:
            self.local_cluster = LocalCluster(n_workers=0, ip=ip)
        except (OSError, IOError):
            self.local_cluster = LocalCluster(n_workers=0, scheduler_port=0,
                                              ip=ip)

        self.packages = list(
            sorted(unique((packages or []) + global_packages, key=first_word)))

        self.knit = Knit(autodetect=autodetect, **kwargs)

        atexit.register(self.stop)

    @property
    def scheduler_address(self):
        return self.local_cluster.scheduler_address

    def start(self, n_workers=1, cpus=1, memory=2048, checks=True,
              **kwargs):
        """
        Initiate workers. If required, environment is first built and uploaded
        to HDFS, and then a YARN application with the required number of
        containers is created.
        
        Parameters
        ----------
        n_workers: int
            How many containers to create
        cpus: int=1
            How many CPU cores is available in each container
        memory: int=2048
            Memory available to each dask worker (in MB)
        checks: bool=True
            Whether to run pre-flight checks before submitting app to YARN
        kwargs: passed to ``Knit.start()``
        
        Returns
        -------
        YARN application ID.
        """
        if self.env is None:
            c = CondaCreator(channels=self.channels, **(self.conda_pars or {}))
            env_name = 'dask-' + sha1(
                '-'.join(self.packages + self.channels).encode()).hexdigest()
            env_path = os.path.join(c.conda_envs, env_name)
            if os.path.exists(env_path + '.zip'):
                # zipfile exists, ready to upload
                self.env = env_path + '.zip'
            elif os.path.exists(env_path):
                # environment exists, can zip and upload
                self.env = zip_path(env_path)
            else:
                # create env from scratch
                self.env = c.create_env(env_name=env_name,
                                        packages=self.packages)
        elif (not self.env.endswith('.zip') and
              not self.env.startswith('hdfs://')):
            # given env directory, so zip it
            self.env = zip_path(self.env)

        # TODO: memory should not be total available?
        bn = os.path.basename(self.env)
        pathsep = '/'  # assume execution is always on posix
        pref = pathsep.join([bn, os.path.splitext(bn)[0]])  # like myenv.zip/myenv
        command = ('{pref}/bin/python {pref}/bin/dask-worker --nprocs=1 '
                   '--nthreads={cpus} --memory-limit={mem} --no-bokeh {addr} '
                   ''.format(cpus=cpus, mem=memory * 1e6, pref=pref,
                             addr=self.local_cluster.scheduler.address))

        files = [self.env] + kwargs.pop('files', [])
        app_id = self.knit.start(command, files=files,
                                 num_containers=n_workers, virtual_cores=cpus,
                                 memory=memory, checks=checks, **kwargs)
        self.app_id = app_id
        return app_id

    def remove_worker(self, container_id):
        """
        Stop worker and remove container

        Parameters
        ----------
        container_id

        Returns
        -------
        None
        """
        self.knit.remove_containers(container_id)

    @property
    def workers(self):
        """
        list of running container ids
        """

        # remove container ...00001 -- this is applicationMaster's container and
        # should not be removed or counted as a worker

        containers = list(self.knit.get_container_statuses())
        containers.sort()
        self.application_master_container = containers.pop(0)
        return containers

    @gen.coroutine
    def _start(self):
        pass

    def stop(self):
        """Kill the YARN application and all workers"""
        if self.knit:
            self.knit.kill()

    def add_workers(self, n_workers=1, cpus=1, memory=2048):
        """
        Non-blocking function to ask Yarn for more containers/dask-workers

        Parameters
        ----------
        n_workers: int
            number of containers to add (default: 1)

        cpus: int
            number of cpus (default: 1)
        memory: int
            amount of memory to allocate per container

        Returns
        -------
        None
        """

        self.knit.add_containers(num_containers=n_workers, virtual_cores=cpus,
                                 memory=memory)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def __repr__(self):
        return "<Dask on YARN cluster {}:{}>".format(
            self.knit.conf['rm'], self.knit.conf['rm_port'])

    def close(self):
        """Stop the scheduler and workers"""
        self.stop()
        self.local_cluster.close()



from __future__ import absolute_import, division, print_function

import sys

if sys.version_info < (3,):
    FileNotFoundError = IOError
    PermissionError = IOError

else:
    PermissionError = PermissionError
    FileNotFoundError = FileNotFoundError

    
try:
    from subprocess import check_output

except ImportError:  # pragma: no cover
    import subprocess
    def check_output(*popenargs, **kwargs):
        """Backported from Python 2.7 as it's implemented as pure python on stdlib."""

        process = subprocess.Popen(stdout=subprocess.PIPE, *popenargs, **kwargs)
        output, unused_err = process.communicate()
        retcode = process.poll()
        if retcode:
            cmd = kwargs.get("args")
            if cmd is None:
                cmd = popenargs[0]
            error = subprocess.CalledProcessError(retcode, cmd)
            error.output = output
            raise error
        return output



from __future__ import absolute_import, division, print_function

import json
import os
import sys
import shutil
import requests
import logging
import tempfile
import zipfile
from subprocess import Popen, PIPE
import warnings

from .exceptions import CondaException
from .utils import shell_out

mini_file = "Miniconda-latest.sh"
miniconda_urls = {"linux": "https://repo.continuum.io/miniconda/"
                           "Miniconda3-latest-Linux-x86_64.sh",
                  "darwin": "https://repo.continuum.io/miniconda/"
                            "Miniconda3-latest-MacOSX-x86_64.sh",
                  "win": "https://repo.continuum.io/miniconda/"
                         "Miniconda3-latest-Windows-x86_64.exe"}

logger = logging.getLogger(__name__)
here = os.path.dirname(__file__)


def miniconda_url():
    """What to download for this platform"""
    if sys.platform.startswith('linux'):
        url = miniconda_urls['linux']
    elif sys.platform.startswith('darwin'):   # pragma: no cover
        url = miniconda_urls['darwin']
    else:  # pragma: no cover
        url = miniconda_urls['win']
    if not sys.maxsize > 2 ** 32:  # pragma: no cover
        # 64bit check
        url = url.replace("_64", "")
    return url


class CondaCreator(object):
    """
    Create Conda Env

    The parameters below can generally be guessed from the system.
    If `conda info` is required, it will only be run on the first
    invocation, and the result cached.

    Parameters
    ----------
    conda_root: str
        Location of a conda installation. The conda executable is expected
        at /bin/conda within.
        If None, runs `conda info` to find out relevant information.
        If no conda is found for `conda info`, or no conda executable exists
        within the given location, will download and install miniconda
        at that location. If the value is None, that location is within the
        source tree.
    conda_envs: str
        directory in which to create environments; usually within the
        source directory (so as not to pollute normal usage of conda)
    miniconda_url: str
        location to download miniconda from, if needed. Uses `miniconda_urls`
        for the appropriate platform if not given.
    channels: list of str
        Channels to specify to conda. Note that defaults in .condarc will also
        be included
    conda_pkgs: str
        Directory containing cached conda packages, normally within conda_root.
    """
    conda_info = {}

    def __init__(self, conda_root=None, conda_envs=None, miniconda_url=None,
                 channels=None, conda_pkgs=None):
        if conda_root is None:
            self._get_conda_info()
            if self.conda_info:
                self.conda_root = self.conda_info['conda_prefix']
            else:
                self.conda_root = os.path.join(here, 'tmp_conda')
        else:
            self.conda_root = conda_root
        self.conda_bin = os.path.join(self.conda_root, 'bin', 'conda')
        if not os.path.exists(self.conda_bin):
            self._install_miniconda(self.conda_root, miniconda_url)
        self.conda_envs = conda_envs or os.sep.join([here, 'tmp_conda', 'envs'])
        self.conda_pkgs = conda_pkgs
        self.channels = channels or []

    def _install_miniconda(self, root, url):
        if os.path.exists(root):
            # conda wants to create the dir - plus this errors if location
            # is not empty
            os.rmdir(root)
        url = url or miniconda_url()
        tmp = tempfile.mkdtemp()
        minifile = os.path.join(tmp, 'Miniconda3')
        logger.debug("Downloading latest Miniconda.sh")
        r = requests.get(url, stream=True)
        with open(minifile, 'wb') as f:
            for chunk in r.iter_content(chunk_size=1024):
                if chunk:
                    f.write(chunk)
            f.flush()
        install_cmd = "bash {0} -b -p {1}".format(minifile, root).split()
        logger.debug("Installing Miniconda in {0}".format(root))
        proc = Popen(install_cmd, stdout=PIPE, stderr=PIPE)
        out, err = proc.communicate()
        logger.debug(out)
        logger.debug(err)
        self.conda_info['conda_prefix'] = root

    def _get_conda_info(self):
        """Ask a conda on PATH where it is installed"""
        if self.conda_info:
            # already did this before
            return
        try:
            self.conda_info.update(json.loads(shell_out(
                ['conda', 'info', '--json'])))
        except (OSError, IOError):
            warnings.warn('No conda found on PATH')

    def _create_env(self, env_name, packages=None, remove=False):
        """
        Create Conda env environment. If env_name is found in self.conda_envs,
        if will be used without checking the existence of any given packages

        Parameters
        ----------
        env_name : str
        packages : list
        remove : bool
            remove environment should it exist - start from 

        Returns
        -------
        path : str
            path to newly created conda environment
        """
        env_path = os.path.join(self.conda_envs, env_name)

        if os.path.exists(env_path):
            if not remove:
                # assume env is OK, ignore packages.
                return env_path
            shutil.rmtree(env_path)

        if not isinstance(packages, list):
            raise TypeError("Packages must be a list of strings")

        ch = []
        [ch.extend(['-c', c]) for c in self.channels]
        cmd = [self.conda_bin, 'create', '-p', env_path, '-y',
               '-q'] + packages + ch
        logger.info("Creating new env {0}".format(env_name))
        logger.info(' '.join(cmd))

        env = dict(os.environ)
        if self.conda_pkgs:
            env['CONDA_PKGS_DIRS'] = self.conda_pkgs
        proc = Popen(cmd, stdout=PIPE, stderr=PIPE, env=env)
        out, err = proc.communicate()

        logger.debug(out)
        logger.debug(err)

        env_python = os.path.join(env_path, 'bin', 'python')

        if not os.path.exists(env_python):
            raise CondaException("Failed to create Python binary at %s."
                                 "" % env_python)

        return env_path

    def find_env(self, env_name):
        """
        Find full path to env_name

        Parameters
        ----------
        env_name : str

        Returns
        -------
        path : str
            path to conda environment
        """

        env_path = os.path.join(self.conda_envs, env_name)

        if os.path.exists(env_path):
            return env_path

    def create_env(self, env_name, packages=None, remove=False):
        """
        Create zipped directory of a conda environment

        Parameters
        ----------
        env_name : str
        packages : list
        remove : bool
            remove environment should it exist

        Returns
        -------
        path : str
            path to zipped conda environment
        """

        if not packages:
            env_path = self.find_env(env_name)
        else:
            env_path = self._create_env(env_name, packages, remove)

        return zip_path(env_path)


def zip_path(path, out_file=None):
    """
    Zip up directory

    Parameters
    ----------
    path : string
    out_path : str
        Output zip-file; if note given, same as input path with
        .zip appended

    Returns
    -------
    path : string
        path to zipped file
    """

    fname = os.path.basename(path) + '.zip'
    env_dir = os.path.dirname(path)
    zFile = out_file or os.path.join(env_dir, fname)

    # ZipFile does not have a contextmanager in Python 2.6
    f = zipfile.ZipFile(zFile, 'w', allowZip64=True)
    try:
        logger.info('Creating: %s' % zFile)
        for root, dirs, files in os.walk(path):
            for file in files:
                relfile = os.path.join(
                    os.path.relpath(root, env_dir), file)
                absfile = os.path.join(root, file)
                try:
                    os.stat(absfile)
                except OSError:
                    logger.info('Skipping zip for %s' % absfile)
                    continue
                f.write(absfile, relfile)
        return zFile
    finally:
        f.close()
